
/* $Header: /cvsroot/faces/faces/faces/imap.c,v 1.1.1.1 2002/04/15 20:20:48 richburridge Exp $
 *
 * This module is a massively modified version of imap_lib.c
 * found in the mutt mailreader.  The original copyrights read:
 *
 * Copyright (C) 1999-2000 Brendan Cully <brendan@kublai.com>
 * 
 * Mutt is distributed under the FSF GNU General Public Licence.  The
 * complete Mutt sources and program can be found at http://www.mutt.org/.
 *
 * This version is
 *  Copyright (C) 2001-2002  Robert Adams (misterblue@misterblue.com)
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 * 
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* If not doing IMAP processing, ifdef the whole file out of existance */
#ifdef IMAP

#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <md5.h>

#include "faces.h"
#include "imap.h"
#include "face_socket.h"

#ifdef PDEBUG
#define  P_DEBUG(stmt)  stmt
#else
#define  P_DEBUG(stmt)
#endif /*PDEBUG*/

/* Function prototypes */
int imap_connect(IMAP_DATA *imap_data);
int imap_disconnect(IMAP_DATA *imap_data);
int imap_logout(IMAP_DATA *);
static int imap_authenticate(IMAP_DATA* imap_data);
int imap_auth_cram(IMAP_DATA *);
int imap_auth_user(IMAP_DATA *);
void imap_error(IMAP_DATA *imap_data, char *msg);

int auth_user_lines(int, char *, void *);
int auth_cram_lines(int, char *, void *);
static void hmac_md5 (const char *, char *, unsigned char *);
void to_base64 (unsigned char *, const unsigned char *, size_t, size_t);
int from_base64 (char *, const char *);

int statinfo_lines(int, char*, void*);

/* Global variables */
long g_lmsgs;
long g_lrecent;

#define MD5_BLOCK_LEN  64
#define MD5_DIGEST_LEN 16

#define SKIPWS(c) while (*(c) && isspace ((unsigned char) *(c))) c++;


/* Create and initialize a structure to hold the state of the IMAP
 * connection.  This lives over several requests so we can remember
 * how full the post box was last time
 */

IMAP_DATA*
build_imapdata(char* host, int port, char* acct, char* passwd)
{
    IMAP_DATA *imapdata;

    P_DEBUG(FPRINTF(stderr, "Build_imapdata(%s, %d, %s\n", host, port, acct));
    imapdata = (IMAP_DATA *) Malloc(sizeof(IMAP_DATA));
    CLEARMEM(imapdata, sizeof(IMAP_DATA));
    imapdata->conn = build_connection(host, port);

    /* make sure both account and passwd have a string in them */
    if (acct) {
        imapdata->account = strdup(acct);
    } else {
        imapdata->account = strdup("");;
    }
    if (passwd) {
        imapdata->passwd = strdup(passwd);
    } else {
        imapdata->passwd = strdup("");
    }
    imapdata->status = IMAP_NONE;
    imapdata->seqno = 1;
    imapdata->size = 0;

    return(imapdata);
}


/* release the IMAP_DATA structure and free what's in it */

int
free_imapdata(IMAP_DATA *imapdata)
{
    P_DEBUG(FPRINTF(stderr, "free_imapdata\n"));
    if (imapdata) {
        if (imapdata->conn) {
            face_socket_close(imapdata->conn);
            free_connection(imapdata->conn);
            imapdata->conn = NULL;
        }
        if (imapdata->account) {
            FREE(imapdata->account);
            imapdata->account = NULL;
        }
        if (imapdata->passwd) {
            FREE(imapdata->passwd);
            imapdata->passwd = NULL;
        }
    }

    return(0);
}


/*
 * Open connection and authenticate
 *  0 - successful,
 * -1 - conection lost,
 * -2 - invalid command or execution error,
 * -3 - authentication canceled.
 */

int
imap_open_connection(IMAP_DATA *imap_data)
{
    int ret;

    P_DEBUG(FPRINTF(stderr, "imap_open_connection()\n"));

    /* connect to the mail server */
    ret = imap_connect(imap_data);
    if (ret < 0) {
        sleep(2);
        return(ret);
    }

    /* do the login thing */
    if (imap_data->status != IMAP_CONNECTED) {
        ret = imap_authenticate(imap_data);

        if (ret == -1) {
            goto err_conn;
        }
        if (ret != IMAP_A_SUCCESS) {
            P_DEBUG(FPRINTF(stderr, 
                    "imap_open_connection(): login failed. %d\n", ret));
            return(ret);
        }

        imap_data->status = IMAP_CONNECTED;
        P_DEBUG(FPRINTF(stderr, "imap_open_connection(): login successful\n"));
    }

    return(IMAP_A_SUCCESS);

err_conn:
    imap_data->status = IMAP_DISCONNECTED;
    FPRINTF(stderr, "Server closed connection!\n");
    sleep(2);

    return(-1);
}


/* undo what was done by imap_open_connection */

int
imap_close_connection(IMAP_DATA *imap_data)
{
    int ret;

    P_DEBUG(FPRINTF(stderr, "imap_close_connection()\n"));
    if (imap_data->status == IMAP_CONNECTED) {
        ret = imap_logout(imap_data);
        ret = imap_disconnect(imap_data);
    }
    imap_data->status = IMAP_DISCONNECTED;

    return(0);
}


/*
 * Open connection
 *  0 - successful,
 * -1 - conection lost,
 * -2 - invalid response.
 */

int
imap_connect(IMAP_DATA *imap_data)
{
    char buf[LONG_STRING];

    imap_data->status = IMAP_NONE;
    if (face_socket_open(imap_data->conn) < 0 ||
        face_socket_readln(imap_data->conn, buf, sizeof(buf)) < 0) {
        SPRINTF(buf,"Error connecting to server: %s", imap_data->conn->host);
        return(-1);
    }

    if (strncmp(buf, "* PREAUTH", 9) == 0) {
        /* for some reason the user is trusted -- already authorized */
        P_DEBUG(FPRINTF(stderr, "User is preauthorized: %s\n", buf));
        imap_data->status = IMAP_CONNECTED;
    } else {
        /* did he say OK? */
        if (strncmp(buf, "* OK", 4)) {
            /* not ok... return the error */
            *imap_data->err_msg = '\0';
            imap_error(imap_data, buf);
            P_DEBUG(FPRINTF(stderr, "%s\n", imap_data->err_msg));
            return(-2);
        } else {
            /* we're connected... not authorized but that's next */
            imap_data->status = IMAP_NONAUTHENTICATED;
        }
    }

    return(0);
}


/* undo what was done by imap_connect */

int
imap_disconnect(IMAP_DATA *imap_data)
{
    int ret;

    P_DEBUG(FPRINTF(stderr, "imap_disconnect()\n"));
    if (imap_data->status != IMAP_NONE) {
        ret = face_socket_close(imap_data->conn);
        imap_data->status = IMAP_NONE;
    }

    return(0);
}


/* logout from IMAP server */

int
imap_logout(IMAP_DATA *imapdata)
{
    int ret = 0;

    if (imapdata->status == IMAP_CONNECTED) {
        P_DEBUG((FPRINTF(stderr, "Closing connection to IMAP server...\n")));

        ret = imap_query(imapdata, "LOGOUT", NULL, NULL);
    }

    imapdata->status = IMAP_DISCONNECTED;

    return(0);
}

/* Copy error message to err_msg buffer in the imap_data
 * This will allow upper levels to print out reasonable error msgs.
 */

void
imap_error(IMAP_DATA *imap_data, char *msg)
{
    char *t, *c;

    t = strchr(imap_data->err_msg, '\0');
    c = msg;

    strncpy(t, c, sizeof(imap_data->err_msg) - strlen(imap_data->err_msg));
}


/* Pass the user and password to the IMAP server and see if
 * we get logged in
 */

static int
imap_authenticate(IMAP_DATA *imap_data)
{
    int ret;

    /* try the IMAP sign-in thing */
    ret = imap_auth_cram(imap_data);
    if (ret != IMAP_A_SUCCESS) {
        /* if CRAM didn't work, try the insecure but basic USER/PASSWD */
        ret = imap_auth_user(imap_data);
        /* pass any error back to the caller */
    }

    return(ret);
}


/* Authenticate using USER/PASS */

int
imap_auth_user(IMAP_DATA *imap_data)
{
    char buf[LONG_STRING];
    int ret;

    sprintf(buf, "LOGIN %s %s", imap_data->account, imap_data->passwd);
    ret = imap_query(imap_data, buf, auth_user_lines, imap_data);
    if (ret == IMAP_A_SUCCESS) {
        if (imap_data->status == IMAP_CONNECTED) {
            ret = IMAP_A_SUCCESS;
        } else {
            ret = IMAP_A_FAILURE;
        }
    }

    return(ret);
}


/* Called for the response for the LOGIN command.  Check for OK'ness */

int
auth_user_lines(int icode, char* buf, void* data)
{
    IMAP_DATA *imap_data = (IMAP_DATA *) data;

    /* see if response tag to our login */
    if (icode & IMAP_CMD_TAG) {
        if (icode & IMAP_CMD_OK) {
            /* if tagged response is "OK" then we're logged in */
            P_DEBUG(FPRINTF(stderr, "Login cmd success: %s\n", buf));
            imap_data->status = IMAP_CONNECTED;
        } else {
            P_DEBUG(FPRINTF(stderr, "Login cmd failed: %s\n", buf));
        }
    }

    return(IMAP_A_SUCCESS);
}


/* Authenticate using CRAM -- RFC2195 */

int
imap_auth_cram(IMAP_DATA *imap_data)
{
    char buf[LONG_STRING];
    int ret = IMAP_A_FAILURE;

    P_DEBUG(FPRINTF(stderr, "Authenticating with CRAM-MD5\n"));
    sprintf(buf, "AUTHENTICATE CRAM-MD5");
    ret = imap_query(imap_data, buf, auth_cram_lines, imap_data);
    if (ret == IMAP_A_SUCCESS) {
        if (imap_data->status == IMAP_CONNECTED) {
            ret = IMAP_A_SUCCESS;
        } else {
            ret = IMAP_A_FAILURE;
        }
    }

    return(ret);
}


/* Called for the response to the AUTHENTICATE command.
 * If first time through, get the challange and send the response.
 */

int
auth_cram_lines(int icode, char* buf, void* data)
{
    IMAP_DATA *imap_data = (IMAP_DATA *) data;
    char ibuf[LONG_STRING*2], obuf[LONG_STRING];
    unsigned char hmac_response[MD5_DIGEST_LEN];
    int len;

    /* A 'continue' is the challange from the server.
     * Calculate and send the response.
     */
    if (icode & IMAP_CMD_CONTINUE) {
        /* From RFC 2195:
         * The data encoded in the first ready response contains a 
         * presumptively arbitrary string of random digits, a timestamp, 
         * and the fully-qualified primary host name of the server. The 
         * syntax of the unencoded form must correspond to that of an 
         * RFC 822 'msg-id' [RFC822] as described in [POP3].
         */
        if ((len = from_base64(obuf, buf + 2)) == -1) {
            FPRINTF(stderr, "Error decoding base64 response.\n");
            goto crambail;
        }

        obuf[len] = '\0';
        P_DEBUG(FPRINTF(stderr, "CRAM challenge: '%s'\n", obuf));

        /* The client makes note of the data and then responds with a string
         * consisting of the user name, a space, and a 'digest'. The latter is
         * computed by applying the keyed MD5 algorithm from [KEYED-MD5] 
         * where the key is a shared secret and the digested text is the 
         * timestamp (including angle-brackets).
         * 
         * Note: The user name shouldn't be quoted. Since the digest can't 
         *       contain spaces, there is no ambiguity. Some servers get 
         *       this wrong, we'll work around them when the bug report 
         *       comes in. Until then, we'll remain blissfully RFC-compliant.
         */
        hmac_md5(imap_data->passwd, obuf, hmac_response);

        /* dubious optimisation I saw elsewhere: make the whole string in 
         * one call.
         */
        snprintf(obuf, sizeof (obuf),
          "%s %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
                 imap_data->account,
                 hmac_response[0], hmac_response[1], hmac_response[2], 
                 hmac_response[3], hmac_response[4], hmac_response[5], 
                 hmac_response[6], hmac_response[7], hmac_response[8], 
                 hmac_response[9], hmac_response[10], hmac_response[11],
                 hmac_response[12], hmac_response[13], hmac_response[14], 
                 hmac_response[15]);
        P_DEBUG(FPRINTF(stderr, "CRAM response: '%s'\n", obuf));

        /* Note: ibuf must be long enough to store the base64 encoding of 
         *       obuf, plus the additional debris.
         */

        to_base64((unsigned char *) ibuf, (unsigned char *) obuf, 
                  strlen(obuf), sizeof(ibuf) - 2);
        strncat(ibuf, "\r\n", sizeof(ibuf));
        P_DEBUG(FPRINTF(stderr, "responsee output: '%s', len=%d\n", 
                        ibuf, strlen(ibuf)));
        face_socket_write(imap_data->conn, ibuf, strlen(ibuf));

crambail:;
        /* fall through -- the server will eventually response with a tag */
    } else if (icode & IMAP_CMD_TAG) {
        /* this is the completion ofthe login sequence -- success? */
        if (icode & IMAP_CMD_OK) {
            imap_data->status = IMAP_CONNECTED;
            P_DEBUG(FPRINTF(stderr, "Auth succeeded: %s\n", buf));
        } else {
            P_DEBUG(FPRINTF(stderr, "Auth failed: %s\n", buf));
        }
    }

    return(0);
}


/* hmac_md5: produce CRAM-MD5 challenge response. */

static void 
hmac_md5(const char *password, char *challenge, unsigned char *response)
{
    MD5_CTX ctx;
    unsigned char ipad[MD5_BLOCK_LEN], opad[MD5_BLOCK_LEN];
    unsigned char secret[MD5_BLOCK_LEN+1];
    unsigned char hash_passwd[MD5_DIGEST_LEN];
    unsigned int secret_len, chal_len;
    int i;

    secret_len = strlen(password);
    chal_len = strlen(challenge);

    /* passwords longer than MD5_BLOCK_LEN bytes are substituted with 
     * their MD5 digests.
     */
    if (secret_len > MD5_BLOCK_LEN) {
        MD5Init(&ctx);
        MD5Update(&ctx, (unsigned char *) password, secret_len);
        MD5Final(hash_passwd, &ctx);
        STRNCPY((char *) secret, (char *) hash_passwd, MD5_DIGEST_LEN);
        secret_len = MD5_DIGEST_LEN;
    } else {
        STRNCPY((char *) secret, password, sizeof (secret));
    }

    memset(ipad, 0, sizeof(ipad));
    memset(opad, 0, sizeof(opad));
    memcpy(ipad, secret, secret_len);
    memcpy(opad, secret, secret_len);

    for (i = 0; i < MD5_BLOCK_LEN; i++) {
        ipad[i] ^= 0x36;
        opad[i] ^= 0x5c;
    }

    /* inner hash: challenge and ipadded secret */
    MD5Init(&ctx);
    MD5Update(&ctx, ipad, MD5_BLOCK_LEN);
    MD5Update(&ctx, (unsigned char *) challenge, chal_len);
    MD5Final(response, &ctx);

    /* outer hash: inner hash and opadded secret */
    MD5Init(&ctx);
    MD5Update(&ctx, opad, MD5_BLOCK_LEN);
    MD5Update(&ctx, response, MD5_DIGEST_LEN);
    MD5Final(response, &ctx);
}


/* do a SELECT query on the connection and get the size and count info */
/* Notice that this means that this one must do a get_statinfo after
 * making a connection to get the mailbox selected.
 */

int
imap_get_statinfo(IMAP_DATA * imap_data, long *msgs, long *boxsize)
{
    int ret;
    
    g_lmsgs = g_lrecent = 0L;
    ret = imap_query(imap_data, "EXAMINE INBOX", statinfo_lines, NULL);
    if (ret != 0) {
        return(-1);
    }

    *msgs = g_lmsgs;
    *boxsize = g_lrecent;   /* fake a boxsize (recent/unread will have to do) */

    return(0);    
}


/* Called for each response of the 'select' command from statinfo */

int
statinfo_lines(int icode, char * buf, void* data)
{
    long lTemp = 0L;
    char sTemp[LONG_STRING];

    if (icode & IMAP_CMD_RESPOND) {
        if (sscanf(buf, "* %ld %s", &lTemp, sTemp) == 2) {
            if (strcmp(sTemp, "EXISTS") == 0) {
                g_lmsgs = lTemp;
            } else if (strcmp(sTemp, "RECENT") == 0) {
                g_lrecent = lTemp;
            }
        }
    }

    return(0);
}


/*
 * First we send the command and then we call the user function
 * funct(int, *line, *data)  for each received line.  Some preprocessing
 * of the received line is done and the called routine is passed a
 * code.  The codes are the IMAP_CMD_* codes
 *
 * The last line of the query is also passed to the user function
 * and after that, this routine returns to it's caller.
 *
 * Returned codes for this routine:
 *  0 - successful,
 * -1 - conection lost,
 * -2 - invalid command or execution error,
 * -3 - error in funct(*line, *data)
 */

int
imap_query(IMAP_DATA *imap_data, char *cmd,
           int (*funct) (int, char *, void *), void *data)
{
    char buf[LONG_STRING];
    char *inbuf;
    int ret, chunk, iline = 0;
    int icode;
    size_t lenbuf = 0;

    if ((imap_data->status != IMAP_CONNECTED) && 
        (imap_data->status != IMAP_NONAUTHENTICATED)) {
        return(-1);
    }

    /* Build command line with the sequence number */
    imap_data->seqno++;
    if (imap_data->seqno > 9999) {
        imap_data->seqno = 0;
    }
    snprintf(imap_data->cmdtag, sizeof(imap_data->cmdtag), "a%04d", 
             imap_data->seqno);
    sprintf(buf, "%s %s\r\n", imap_data->cmdtag, cmd);
    ret = face_socket_write(imap_data->conn, buf, strlen(buf));

    if (ret < 0) {
        return(ret);
    }

    /* Allocate the buffer that we will build */
    inbuf = (char*) Malloc(sizeof(buf));

    while (1) {
        /* read one chunk of the input stream */
        chunk = face_socket_readln(imap_data->conn, buf, sizeof(buf));
        if (chunk < 0) {
            imap_data->status = IMAP_DISCONNECTED;
            ret = -1;
            break;
        }

        /* add the read buf to the inbuf we are building */
        STRNCPY(inbuf + lenbuf, buf, sizeof(buf));

        /* is the chunk we read just part of  the line? */
        if (chunk >= sizeof(buf)) {
            /* if just part of the line, setup to append */
            lenbuf += strlen(buf);
            if ((inbuf = (char*) realloc(inbuf, 
                                         lenbuf + sizeof (buf))) == NULL) {
                FPRINTF(stderr, "imap_fetch_data: Out of memory.\n");
            }
        } else {
            /* we have a line from the server -- classify it */
            icode = 0;
            if (strncmp(inbuf, "+ ", 2) == 0) {
                icode = IMAP_CMD_CONTINUE;
            } else if (strncmp(inbuf, "* ", 2) == 0) {
                icode = IMAP_CMD_RESPOND;
                if (strncmp(inbuf, "* OK", 4) == 0) {
                    icode |= IMAP_CMD_OK;
                } else if (strncmp(inbuf, "* BAD", 5)) {
                    icode |= IMAP_CMD_BAD;
                } else if (strncmp(inbuf, "* NO", 4) == 0) {
                    icode |= IMAP_CMD_NO;
                }
            } else if (strncmp(inbuf, imap_data->cmdtag, TAGSIZE) == 0) {
                icode = IMAP_CMD_TAG;
                if (strncmp(&inbuf[TAGSIZE], " OK", 3) == 0) {
                    icode |= IMAP_CMD_OK;
                } else if (strncmp(&inbuf[TAGSIZE], " NO", 3) == 0) {
                    icode |= IMAP_CMD_NO;
                } else if (strncmp(&inbuf[TAGSIZE], " BAD", 4) == 0) {
                    icode |= IMAP_CMD_BAD;
                }
            }

            /* pass the read line to the caller's processing routine */
            iline++;
            if (funct != NULL) {
                if (funct(icode, inbuf, data) < 0) {
                    ret = -3;
                    break;
                }
            }
            lenbuf = 0;
            if (icode & IMAP_CMD_TAG) {
                /* if the TAG code then last line of the query was seen */
                ret = IMAP_A_SUCCESS;        /* looks like it worked*/
                break;
            }
        }
    }

    FREE(inbuf);

    return(ret);
}


/* 
 * This code is heavily modified from fetchmail (also GPL'd, of
 * course) by Brendan Cully <brendan@kublai.com>.
 * 
 * Original copyright notice:
 * 
 * The code in the fetchmail distribution is Copyright 1997 by Eric
 * S. Raymond.  Portions are also copyrighted by Carl Harris, 1993
 * and 1995.  Copyright retained for the purpose of protecting free
 * redistribution of source. 
 */

#define BAD -1

char B64Chars[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '+', '/'
};

int Index_64[128] = {
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,62, -1,-1,-1,63,
    52,53,54,55, 56,57,58,59, 60,61,-1,-1, -1,-1,-1,-1,
    -1, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,-1, -1,-1,-1,-1,
    -1,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,-1, -1,-1,-1,-1
};

#define base64val(c) Index_64[(unsigned int)(c)]


/* Raw bytes to null-terminated base 64 string */

void 
to_base64(unsigned char *out, const unsigned char *in, size_t len, size_t olen)
{
    while (len >= 3 && olen > 10) {
        *out++ = B64Chars[in[0] >> 2];
        *out++ = B64Chars[((in[0] << 4) & 0x30) | (in[1] >> 4)];
        *out++ = B64Chars[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
        *out++ = B64Chars[in[2] & 0x3f];
        olen  -= 4;
        len   -= 3;
        in    += 3;
    }

    /* clean up remainder */
    if (len > 0 && olen > 4) {
        unsigned char fragment;

        *out++ = B64Chars[in[0] >> 2];
        fragment = (in[0] << 4) & 0x30;
        if (len > 1) {
            fragment |= in[1] >> 4;
        }
        *out++ = B64Chars[fragment];
        *out++ = (len < 2) ? '=' : B64Chars[(in[1] << 2) & 0x3c];
        *out++ = '=';
    }
    *out = '\0';
}


/* Convert '\0'-terminated base 64 string to raw bytes.
 * Returns length of returned buffer, or -1 on error.
 */

int 
from_base64(char *out, const char *in)
{
    int len = 0;
    unsigned char digit1, digit2, digit3, digit4;

    do {
        digit1 = in[0];
        if (digit1 > 127 || base64val (digit1) == BAD) {
            return(-1);
        }
        digit2 = in[1];
        if (digit2 > 127 || base64val (digit2) == BAD) {
            return(-1);
        }
        digit3 = in[2];
        if (digit3 > 127 || ((digit3 != '=') && (base64val (digit3) == BAD))) {
            return(-1);
        }
        digit4 = in[3];
        if (digit4 > 127 || ((digit4 != '=') && (base64val (digit4) == BAD))) {
            return(-1);
        }
        in += 4;

        /* digits are already sanity-checked */
        *out++ = (base64val(digit1) << 2) | (base64val(digit2) >> 4);
        len++;
        if (digit3 != '=') {
            *out++ = ((base64val(digit2) << 4) & 0xf0) | 
                      (base64val(digit3) >> 2);
            len++;
            if (digit4 != '=') {
                *out++ = ((base64val(digit3) << 6) & 0xc0) | base64val(digit4);
                len++;
            }
        }
    } while (*in && digit4 != '=');

    return(len);
}

#endif /*IMAP*/
